/**
 * @file dns.c
 * \author Michal Kovacik <ikovacik@fit.vutbr.cz>#
 * \date 02.02.2014
 * \brief Process plugin for parsing DNS headers on input in 
 * flowmonexp 4.0.0 and later
 *
 * Copyright (C) 2014 CESNET
 *
 * LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL) version 2 or later, in which case the provisions
 * of the GPL apply INSTEAD OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id$
 *
 */

/* INCLUDES */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include <arpa/inet.h>

#include "dns.h"
#include <flowmonexp/plugin_process.h>


/* DEFINES */
/** Print when verbose mode is on*/
#define PRINTV(format, args...) do {if (VERBOSE) { fprintf(stdout, format, ##args); fflush(stdout); } } while (0)

/** Print when debug mode is on*/
#define PRINTD(format, args...) do {if (DBG) { fprintf(stdout, format, ##args); fflush(stdout); } } while (0)

/** Print when debug of verbose mode is on*/
#define PRINTA(format, args...) do {if (VERBOSE || DBG) { fprintf(stdout, format, ##args); fflush(stdout); } } while (0)

/** VERSION CONSTANT */
FLOWMON_API_VERSION;

/* GLOBAL */
/** Set plugin type for flowmonexp*/
SET_PLUGIN_TYPE(PLUGIN_TYPE_PROCESS);

/** Verbose mode flag*/
bool VERBOSE;

/** Debug mode flag*/
bool DBG;

/** Parse all DNS fields flag */
bool ALL_DNS;


/**
 * Structure representing plugin state
 */
typedef struct {

	int data;	/* offset of data */
	unsigned char *buffer;	/* allocated buffer */

} plugin_private_dns;

/**
 * \brief Plugin description
 */
static plugin_desc_t pcap_desc = {
	"dns",
	"Plugin for parsing DNS traffic on input\n"
	    "Parameters:\n"
	    " debug=y/n		Switch on/off debug mode (default: same as flowmonexp)\n"
	    " verbose=y/n	Switch on/off verbose mode (default: same as flowmonexp)\n"
	    " all=y/n		Switch on/off parsing of ALL of present DNS fields\n",
	sizeof(dns_record_t),
	PLUGIN_NEED_FULL_PACKET_INPUT
};

/**
 * \brief Returns plugin description
 * @return Plugin Description
 */
PLUGIN_PROCESS_DESC {
	return &pcap_desc;
};


/**
 * \brief Function for parsing of the input parameters
 * @param *params pointer to parameter string
 * @param *conf plugin structure to fill with parameters
 * @return 0 when ok, non-zero if not
 */
int parse_params(char *params, plugin_private_dns *conf) {
	char *token, *value, *param;
	/* expect parameter or expect value */
	uint8_t state = 0;
	/* sep */
	token = strtok(params, ",=");
	
	/* while something on input */
	while (token != NULL) {
		if (state == 0) {
			param = token;
		} else {
			value = token;
			PRINTA("param: %s\t", param);
			PRINTA("value: %s\n", value);
			if (strcmp("verbose", param) == 0) {
				if (strcmp(value, "y") == 0) {
					VERBOSE = true;
				} else if (strcmp(value, "n") == 0) {
					VERBOSE = false;
				} else {
					fprintf(stderr,	" value of 'verbose' parameter is not valid");
				}
			} else if (strcmp("debug", param) == 0) {
				if (strcmp(value, "y") == 0) {
					DBG = true;
				} else if (strcmp(value, "n") == 0) {
					DBG = false;
				} else {
					fprintf(stderr,	" value of 'debug' parameter is not valid");
				}
#ifdef PARSE_ALL
			} else if (strcmp("all", param) == 0) {
				/* parse all DNS fields */
				if (strcmp(value, "y") == 0) {
					ALL_DNS = true;
				} else if (strcmp(value, "n") == 0) {
					ALL_DNS = false;
				} else {
					fprintf(stderr,	" value of 'all' parameter is not valid");
				}
#endif
			} else {
				fprintf(stderr, "Unknown parameter '%s'\n",
					param);
				return 1;
			}
		}

		/* next param */
		token = strtok(NULL, ",=");
		state = 1 - state;
	}

	return 0;
}


/**
 * \brief Function for parsing hostname (called for query and response, or for any hostname present in response data)
 * @param *buf buffer of data to parse
 * @param buf_length buffer size
 * @param *off offset
 * @param depth depth in dns structure
 * @param n hostname
 * @return return -1 on error, 0 on ok
 */
static inline int parse_name (unsigned char *buf, int buf_length, size_t *off, unsigned int depth, char *n) {
	const uint8_t *input = (const uint8_t *) buf;	/* 8bit input from buffer */
	char xstr[NAME_SIZE];	/* hostname part for compressed hostname in response */
	char current_level[LEVEL_SIZE + 1];		/* parsed domain level */
	
	*n = '\0';
	*current_level = '\0';
	
	uint8_t len;	/* length of hostname part */
	size_t xoff;	/* offset for compressed hostname in response */
	
	bool cutLast = false;	/* flag deciding if last character have to be cut */
	
	while (1) {	/* while the loop is not broken */
		
		/* out of buffer */
		if (*off >= buf_length) {
			return ERR;
		}
		
		/* get first part length */
		len = input[*off];
		
		/* end of hostname, finish parsing */
		if (0 == len) {
			(*off)++;
			break;
		}
		
		/* hostname present - not compressed */
		if (len < DNS_NAME_UNCOMPRESSED) {

			/* test buffer bounds */
			if (*off + 1 + len > buf_length) {
				return ERR;
			}
			
			/* add new part */
			if (len < LEVEL_SIZE) {
			    memcpy(current_level, &buf[*off + 1], len);
			    current_level[len] = '.';
			    current_level[len + 1] = '\0';
			} else { 
			    memcpy(current_level, &buf[*off + 1], LEVEL_SIZE - 1);
			    current_level[LEVEL_SIZE - 1] = '.';
			    current_level[LEVEL_SIZE] = '\0';
			}
			
			/* assign new name */
			if (strlen(n) < NAME_SIZE) {
			    if ((len + strlen(n)) < (NAME_SIZE-1)) {
				strcat(n, current_level);
			    } else {
				strncat(n, current_level, (NAME_SIZE - strlen(n) - 1));
			    }
			}
			
			/* move offset to length of next part of hostname */
			*off += 1 + len;
		  
		} else if ((len & DNS_NAME_COMPRESSED) == DNS_NAME_COMPRESSED ) {
		  /* compressed hostname in response - pointer is present */
		
			/* highest depth of jumping boundary, prevents infinite cycle */
			if (depth > DNS_MAXDEPTH) {
				return ERR;
			}

			/* check buffer boundaries */
			if (*off + 1 > buf_length) {
				return ERR;
			}

			/* offset of real hostname in query */
			xoff = ((len - DNS_NAME_COMPRESSED) << 8) + input[*off + 1];
			
			/* call recursivelly parsing of uncompressed name */
			if (ERR == parse_name (buf, buf_length, &xoff, depth + 1, xstr))
				return ERR;

			/* length of first part and compressed part */
			size_t len1 = strlen(n);
			size_t len2 = strlen(xstr);

			/* append compressed part */
			if ((len1 + len2) < NAME_SIZE) {
			    strcat(n, xstr);
			} else {
			    strncat(n, xstr, (NAME_SIZE - len1 - 1));
			}

			/* move offset after compressed part */
			*off += 2;
			
			/* name doesnt end with ".", doesnt need to be cut away */
			cutLast = true;
		
			break;
		} else {
			/* unsupported functionality */
			return ERR;
		}
	}
       
	/* cut the last "." if needed - uncompressed names*/
	if (!cutLast) {
		if (0 < strlen(n))
			n[strlen(n)-1] = '\0';
	}

	return OK;
}


/**
 * \brief Function for parsing DNS query 
 * @param *buf buffer of data to parse
 * @param buf_length buffer size
 * @param *off offset
 * @param *dns dns_record_t structure
 * @return -1 on error, 0 on OK
 */
static inline int parse_query (unsigned char *buf, int buf_length, size_t *off, dns_record_t *dns) {
	
	query_line_t ql;	/* structure of query header size */
	query_t q;		/* full dns query structure */
	memset(&q, 0, sizeof(query_t));
	
	/* parse query name */
	if (ERR == parse_name (buf, buf_length, off, 0, q.name))
		return ERR;
	
	strcpy(dns->name, q.name);
	
	/* check validity of packet by buffer length */
	if (*off + sizeof (query_line_t) > buf_length) {
		return ERR;
	}

	/* fill query header structure */
	memcpy (&ql, &buf[*off], sizeof(query_line_t));

	/* move offset behind query header */
	*off += sizeof(query_line_t);
	
	/* fill dns query structure variables */
	q.type = ntohs(ql.type);
	q.class = ntohs(ql.class);
	q.qlength = strlen(q.name) + 1;
	
	dns->type = q.type;
	dns->class = q.class;
	
#ifdef PARSE_ALL
	dns->d.q = q;
#endif
	
	PRINTD("Q: %s - ", q.name);

	return OK;
}

/**
 * \brief parse DNS record
 * @param *buf buffer of data to parse
 * @param buf_length buffer size
 * @param *off offset
 * @param *dns dns_record_t structure
 * @return -1 on error, 0 on OK
 */
static inline int parse_record(unsigned char *buf, int buf_length, size_t *off, dns_record_t *dns) {

	uint16_t data_len;	/* length of rdata */	
	rr_line_t rl;		/* structure of response header size */
	record_t r;		/* full response record structure */
	memset(&r, 0, sizeof(record_t));
	
	/* response type formatting structures and temporary variables */
	int ret;		/* return value of name parsing */
	int print_err;		/* return value of usd snprintf */
	size_t new_off;		/* new offset for formatting */
	//char tmp_num[12];	/* temporary variable for itoa conversion */
	struct soa_data_t soa;	/* structure for fixed data */
	uint16_t mxpref;
	struct srv_data_t srv;
	char ndup[NAME_SIZE];
	char *tok;
	struct ds_data_t ds;
	struct rrsig_data_t rrsig;
	struct dnskey_data_t dnskey;
	
	/* parse name from response */
	if (ERR == parse_name(buf, buf_length, off, 0, r.name))
		return ERR;
	
	/* check boundaries for empty response - if so, NO DNS record of that type */
	if (*off + sizeof (rr_line_t) > buf_length) {
		return ERR;
	}
	
	/* fill response data structure */
	memcpy (&rl, &buf[*off], sizeof (rr_line_t));
	
	/* move offset after response header */
	(*off) += sizeof (rl);
	
	/* fill main dns structure */
	r.type = ntohs(rl.type);
	r.class = ntohs(rl.class);
	r.ttl = ntohl (rl.ttl);
	r.rlength = ntohs(rl.data_len);
	data_len = ntohs(rl.data_len);
	
	/* write parsed data to mains tructure */
	dns->ttl = r.ttl;
	dns->rlength = r.rlength;
	
	/* check packet validity */
	if (*off + data_len > buf_length) {
		return ERR;
	}
	
	PRINTD("R: %s %d\n", r.name, r.type);
		
	/* parse exact type of DNS response */
	switch (r.type)
	{
	  /* for NS, CNAME, PTR, DNAME parse only hostname */
	  case DNS_TYPE_NS:
	  case DNS_TYPE_CNAME:
	  case DNS_TYPE_PTR:
	  case DNS_TYPE_DNAME:
	    
	    new_off = *off;
	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.hostname);
	    if ((ERR == ret) || (*off + data_len != new_off)) {
	      return ERR;
	    }
	    
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%s", r.data.hostname);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* MINFO has 2 domain names divided by white space on the output */
	  case DNS_TYPE_MINFO:
	    
	    /* RMAILBX - <domain-name> which specifies mailbox responsible for mailing list or mailbox */
	    new_off = *off;
	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.hostname);
	    if ((ERR == ret) || (*off + data_len != new_off)) {
	      return ERR;
	    }

	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%s", r.data.hostname);
	    
	    int rdata_offset = strlen(r.rdata);
	
	    /* EMAILBX - <domain-name> which specifies mailbox which is to receive errors */
	    new_off = *off;
	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.hostname);
	    if ((ERR == ret) || (*off + data_len != new_off)) {
	      return ERR;
	    }
	
	    /* concatenate domain names with " " */
	    print_err = snprintf(r.rdata+rdata_offset, RESPONSE_SIZE, " %s", r.data.hostname);
	
	    r.slength = strlen(r.rdata) + 1;

	    break;

	  /* SOA fields separated by white space*/
	  case DNS_TYPE_SOA:
	
	    new_off = *off;

	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.soa.mname);
	    /* 1st validity and boundary check */
	    if ((ERR == ret) || (new_off + sizeof (struct soa_data_t) > buf_length)) {
	      return ERR;
	    }
	
	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.soa.rname);
	    if ((ERR == ret) || (new_off + sizeof (struct soa_data_t) > buf_length)) {
	      return ERR;
	    }
	
	    /* fill main soa structure */
	    memcpy (&soa, &buf[new_off], sizeof (struct soa_data_t));
	    r.data.soa.serial = ntohl (soa.serial);
	    r.data.soa.refresh = ntohl (soa.refresh);
	    r.data.soa.retry = ntohl (soa.retry);
	    r.data.soa.expire = ntohl (soa.expire);
	    r.data.soa.min_ttl = ntohl (soa.minimum);
	
	    /* move offset behind payload */
	    (new_off) += sizeof (struct soa_data_t);
	
	    /* validity check */
	    if (*off + data_len != new_off) {
	      return ERR;
	    }
	
	    /* building string with assistance of itoa function */
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%s %s %d %d %d %d %d", 
		r.data.soa.mname, r.data.soa.rname, r.data.soa.serial, r.data.soa.refresh,
		r.data.soa.retry, r.data.soa.expire, r.data.soa.min_ttl);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;

	  /* MX output is priority and hostname separated by white space */
	  case DNS_TYPE_MX:
	
	    new_off = *off;
	    /* validity check for mx priority */
	    if (*off + sizeof (uint16_t) > buf_length) {
	      return ERR;
	    }
	
	    /* fill priority */
	    memcpy (&mxpref, &buf[*off], sizeof (uint16_t));
	
	    /* parse mx hostname */
	    new_off += sizeof (uint16_t);
	
	    r.data.mx.priority = ntohs(mxpref);
	
	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.mx.mxhost);
	
	    /* check validity */
	    if ((ERR == ret) || (*off + data_len != new_off)) {
	      return ERR;
	    }
	
	    /* build output string */
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%d %s", r.data.mx.priority, r.data.mx.mxhost);
	
	    r.slength = strlen(r.rdata) + 1;

	    break;
	
	  /* A record output is formatted in IPv4 format */
	  case DNS_TYPE_A:
	    /* validity check */
	    if (*off + data_len > buf_length) {
	      return ERR;
	    }
	
	    /* get ip */
	    unsigned long tmp_a;
	    memcpy (&tmp_a, &buf[*off], sizeof (unsigned long));
	
	    /* convert ip to IP format */
	    inet_ntop(AF_INET, &(tmp_a), r.rdata, INET_ADDRSTRLEN);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* AAAA record output is formatted in IPv6 format */
	  case DNS_TYPE_AAAA:
	    if (*off + data_len > buf_length) {
	      return ERR;
	    }
	
	    /* get ip */
	    unsigned char tmp_aaaa[DNS_IP6CHARS];
	    memcpy (&tmp_aaaa, &buf[*off], DNS_IP6CHARS*sizeof(unsigned char));
	
	    /* convert ip to IP format */
	    inet_ntop(AF_INET6, &(tmp_aaaa), r.rdata, INET6_ADDRSTRLEN);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* SRV fields separated by white space */
	  case DNS_TYPE_SRV:
	
	    /* valid hostnames must start with "_" */
	    if ('_' != r.name[0]) {
	      return ERR;
	    }
	
	    /* necessary string from "._$PROTO" not present */
	    if (NULL == strstr(r.name, "._")) {
	      return ERR;
	    }
	
	    new_off = *off;
	
	    /* check boundaries */
	    if (*off + sizeof (struct srv_data_t) > buf_length) {
	      return ERR;
	    }
	
	    /* parse fixed length data */
	    memcpy (&srv, &buf[*off], sizeof (struct srv_data_t));
	    new_off += sizeof (struct srv_data_t);

	    r.data.srv.priority = ntohs (srv.prio);
	    r.data.srv.weight = ntohs (srv.weight);
	    r.data.srv.port = ntohs (srv.port);

	    /* parse hostname into components: "_$SERVICE._$PROTO.$DOMAIN_NAME" */
	    strcpy(ndup, r.name);
	    tok = strtok (ndup, ".");
	    strcpy(r.data.srv.service, &tok[1]);
	    tok = strtok (NULL, ".");
	
	    if ((NULL == tok)) {
	        return ERR;
	    }

	    strcpy(r.data.srv.protocol, &tok[1]);
	    tok = strtok (NULL, ".");
	
	    if (NULL == tok)
		return ERR;

	    strcpy(r.data.srv.name, tok);

	    /* parse target hostname */
	    ret = parse_name (buf, buf_length, &new_off, 0, r.data.srv.target);
	
	    if ((ret == ERR) || (*off + data_len != new_off)) {
	      return ERR;
	    }

	    /* formatted output */
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%s %s %s %s %d %d %d", 
		r.data.srv.service, r.data.srv.protocol, r.data.srv.name, r.data.srv.target,
		r.data.srv.priority, r.data.srv.weight, r.data.srv.port);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* DS fields separated by white space */
	  case DNS_TYPE_DS:
	
	    new_off = *off;
	
	    /* check boundaries */
	    if (*off + sizeof (struct ds_data_t) > buf_length) {
	      return ERR;
	    } 
	
	    /* parse fixed length data */
	    memcpy (&ds, &buf[*off], sizeof(struct ds_data_t));
	
	    r.data.ds.keytag = ntohs (ds.keytag);
	    r.data.ds.algorithm = ds.alg;
	    r.data.ds.digest = ds.digest;
	
	    /* parse public key */
	    /*
	    if (data_len < RESPONSE_SIZE) {
		memcpy (r.data.ds.publickey, &buf[*off], data_len - sizeof(struct ds_data_t));
		r.data.ds.publickey[sizeof(struct ds_data_t) - 1] = '\0';
	    }
	    else {
		memcpy (r.data.ds.publickey, &buf[*off], RESPONSE_SIZE - sizeof(struct ds_data_t));
		r.data.ds.publickey[sizeof(struct ds_data_t) - 1] = '\0';
	    }
	    */
	
	    /* check validity */
	    if (new_off + data_len != *off + (data_len)) {
	      return ERR;
	    }
	
	    /* build output string, without binary key */
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%d %d %d <key>", 
		r.data.ds.keytag, r.data.ds.algorithm, r.data.ds.digest);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* RRSIG fields separated by white space */
	  case DNS_TYPE_RRSIG:
	
	    new_off = *off;
	
	    /* check boundaries */
	    if (*off + sizeof (struct rrsig_data_t) > buf_length) {
	      return ERR;
	    }
	
	    /* parse fixed length data */
	    memcpy (&rrsig, &buf[*off], sizeof(struct rrsig_data_t));
	    r.data.rrsig.type_covered = ntohs (rrsig.type);
	    r.data.rrsig.algorithm = rrsig.alg;
	    r.data.rrsig.labels = rrsig.labels;
	    r.data.rrsig.original_ttl = ntohl (rrsig.ttl);
	    r.data.rrsig.signature_expiration = ntohl (rrsig.sig_exp);
	    r.data.rrsig.signature_inception = ntohl (rrsig.sig_inc);
	    r.data.rrsig.keytag = ntohs (rrsig.keytag);
	
	    /* parse public key */
	    /*
	    if (data_len < RESPONSE_SIZE) {
		memcpy (r.data.rrsig.signer_signature, &buf[*off], data_len - sizeof(struct rrsig_data_t));
		r.data.rrsig.signer_signature[sizeof(struct rrsig_data_t) - 1] = '\0';
	    } else {
		memcpy (r.data.rrsig.signer_signature, &buf[*off], RESPONSE_SIZE - sizeof(struct rrsig_data_t));
		r.data.rrsig.signer_signature[sizeof(struct rrsig_data_t) - 1] = '\0';
	    }
	    */
	
	    /* check validity */
	    if (new_off + data_len != *off + (data_len)) {
	      return ERR;
	    }
	
	    /* build output string, without binary signer signature */
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%d %d %d %d %d %d %d <key>", 
		r.data.rrsig.type_covered, r.data.rrsig.algorithm, r.data.rrsig.labels, r.data.rrsig.original_ttl,
		r.data.rrsig.signature_expiration, r.data.rrsig.signature_inception, r.data.rrsig.keytag);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* DNSKEY fields separated by white space */
	  case DNS_TYPE_DNSKEY:
	
	    new_off = *off;
	
	    /* check boundaries */
	    if (*off + sizeof (struct dnskey_data_t) > buf_length) {
	      return ERR;
	    }
	
	    /* parse fixed length data */
	    memcpy (&dnskey, &buf[*off], sizeof(struct dnskey_data_t));
	    r.data.dnskey.flags = ntohs (dnskey.flags);
	    r.data.dnskey.protocol = dnskey.proto;
	    r.data.dnskey.algorithm = dnskey.alg;
	
	    /* parse public key */
	    /*
	    if (data_len < RESPONSE_SIZE) {
		memcpy (r.data.dnskey.publickey, &buf[*off], data_len - sizeof(struct dnskey_data_t));
		r.data.dnskey.publickey[sizeof(struct dnskey_data_t) - 1] = '\0';
	    } else {
		memcpy (r.data.dnskey.publickey, &buf[*off], RESPONSE_SIZE - sizeof(struct dnskey_data_t));
		r.data.dnskey.publickey[sizeof(struct dnskey_data_t) - 1] = '\0';
	    }
	    */
	
	    /* check validity */
	    if (new_off + data_len != *off + (data_len)) {
	      return ERR;
	    }
	
	    /* build output string, without binary public key */
	    print_err = snprintf(r.rdata, RESPONSE_SIZE, "%d %d %d <key>", 
		r.data.dnskey.flags, r.data.dnskey.protocol, r.data.dnskey.algorithm);
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* TXT record output is as plain text with length in front of it*/
	  case DNS_TYPE_TXT:
	    {
		size_t xoff = *off; //use temporary offset
		uint16_t l = 0; //all TXT lengths summary
		
		unsigned int cond = data_len;
		if (data_len > RESPONSE_SIZE)
		    cond = RESPONSE_SIZE;
		
		 while (cond > l) {
		    uint8_t len; //one length
		    char tmp[RESPONSE_SIZE]; //temporary string
		    memset(tmp, 0, RESPONSE_SIZE);
		    const uint8_t *input = (const uint8_t *) buf;
		    len = input[xoff++]; //get length and skip it
		
		    //space before next part of text
		    if (l != 0) {
			strcat(r.rdata, " ");
		    }
		
		    //check if length is valid
		    if ((len > 0) && (len <= (data_len-l))) {
		
			//RDATA length is shorter than limit
			if (data_len < (RESPONSE_SIZE - l)) {
			    memcpy(tmp, &buf[xoff], len*sizeof(char));
			    tmp[len] = '\0';
			} else {
			//RDATA is longer than limit, actual length is
			//longer than limit, or actual length is smaller
			    if ((l + len) >= RESPONSE_SIZE) {
				memcpy(tmp, &buf[xoff], RESPONSE_SIZE - l - 1);
				tmp[RESPONSE_SIZE-1] = '\0';
			    } else {
				memcpy(tmp, &buf[xoff], len);
				tmp[len] = '\0';
			    }
			}
		
			strcat(r.rdata, tmp); //append next string
		    }
		
		    l += len + 1;	//sum of TXT lengths
		    xoff += len;	//set new temporary offset
		}
		
		r.slength = strlen(r.rdata);
		
		break;
	    }
	  /* HINFO, ISDN record output is as plain text */
	  case DNS_TYPE_HINFO:
	  case DNS_TYPE_ISDN:
	
	    if (data_len < RESPONSE_SIZE) {
		memcpy(r.rdata, &buf[*off], data_len*sizeof(char));
		r.rdata[data_len] = '\0';
	    } else {
		memcpy(r.rdata, &buf[*off], RESPONSE_SIZE);
		r.rdata[RESPONSE_SIZE-1] = '\0';
	    }
	
	    r.slength = strlen(r.rdata) + 1;
	
	    break;
	
	  /* default output is in RAW format, dont print it for now */
	  default:
	
	    /*
	    r.data.raw.data_len = data_len;
	
	    if (data_len < RESPONSE_SIZE) {
		memcpy(r.data.raw.data, &buf[*off], data_len*sizeof(char));
		r.data.raw.data[data_len] = '\0';
	    } else {
		memcpy(r.data.raw.data, &buf[*off], RESPONSE_SIZE*sizeof(char));
		r.data.raw.data[RESPONSE_SIZE-1] = '\0';
	    }
	
	    strcpy (r.rdata, r.data.raw.data);
	    r.slength = r.rlength;
	    */
	
	    strncpy (r.rdata, "(not_impl)", 11);
	    r.slength = 11;
	
	    break;
	}
	
	/* move offset after response part */
	(*off) += data_len;
	
	/* store parsed data to main dns structure */
	strcpy(dns->rdata, r.rdata);
	
#ifdef PARSE_ALL
	dns->d.r = r;
#endif
	
	PRINTD("R-DATA: %s\n", r.rdata);
	
	return OK;
}


/**
 * \brief Function for skipping parts of packet that are not to be parsed
 * @param *buf buffer of data to parse
 * @param buf_length buffer size
 * @param *off offset
 * @return -1 on error, 0 on OK
 */
static inline int parse_skip(unsigned char *buf, int buf_length, size_t *off) {
	
	uint16_t data_len;	/* length of rdata */
	rr_line_t rl;		/* structure of response header size */
	record_t r;		/* full response structure */
	
	/* parse name from response */
	if (ERR == parse_name(buf, buf_length, off, 0, r.name))
		return ERR;
	
	/* check boundaries for empty response - if so, NO DNS record of that type */
	if (*off + sizeof (rr_line_t) > buf_length) {
		//PRINTD("X %zu %lu %d %s\n", *off, sizeof(rr_line_t), buf_length, xx);
		return ERR;
	}
	
	/* fill response data structure */
	memcpy (&rl, &buf[*off], sizeof (rr_line_t));
	data_len = ntohs(rl.data_len);
	/* move offset after response header */
	(*off) += sizeof(rr_line_t);
	
	/* check packet validity */
	if (*off + data_len > buf_length) {
		return ERR;
	}
	
	/* move offset after response part */
	(*off) += data_len;
	
	return OK;
}


/**
 * \brief parse additional section ofDNS record
 * @param *buf buffer of data to parse
 * @param buf_length buffer size
 * @param *off offset
 * @param *dns dns_record_t structure
 * @return -1 on error, 0 on OK
 */
static inline int parse_additional(unsigned char *buf, int buf_length, size_t *off, dns_record_t *dns) {

	uint16_t data_len;	/* length of rdata */	
	rr_line_t rl;		/* structure of response header size */
	additional_t a;		/* full additional section structure */
	
	/* parse name from response */
	if (ERR == parse_name(buf, buf_length, off, 0, a.name))
		return ERR;
	
	/* fill response data structure */
	memcpy (&rl, &buf[*off], sizeof (rr_line_t));
	
	/* move offset after response header */
	(*off) += sizeof(rl);
	
	/* check packet validity */
	if (*off + ntohs(rl.data_len) > buf_length) {
		return ERR;
	}
	
	/* parse exact type of DNS aditional response */
	switch (ntohs(rl.type))
	{
	  /* for OPT type start parsing */
	  case DNS_TYPE_OPT:
	    /* fill main dns structure */
	    a.type = ntohs(rl.type);
	    a.payload_size = ntohs(rl.class);
	    a.flags = rl.ttl;
	    a.rlength = ntohs(rl.data_len);
	    data_len = ntohs(rl.data_len);
	
	    uint16_t lo = (a.flags);
	    uint16_t hi = (a.flags >> 16);
	
	    /* parse flags */
	    a.ext_rcode = lo ;
	    a.version = lo >> 8;
	    a.do_bit = ((ntohl(hi) & 0x80000000) == 0x80000000) ? 0x1 : 0x0;
	    a.z_bit = hi >> 1;
	    
	    dns->payload_size = a.payload_size;
	    dns->do_bit = a.do_bit;
	    a.add_valid = 1;
	    
#ifdef PARSE_ALL
	    dns->d.a = a;
#endif
	    
	    PRINTD("A: %d %d \n", a.payload_size, a.do_bit);
	    break;
	    
	  /* default type */
	  default:
	
	    break;
	}
	
	/* move offset after response part */
	(*off) += ntohs(rl.data_len);
	
	return OK;
}


/**
 * \brief Function for parsing DNS header, called first to fill dns_record_t using parse_query() and parse_record()
 * @param buf packet data buffer
 * @param dns structure for dns data
 * @param buflen length of packet data buffer
 * @return 0 for dns headers, non 0 for fail or unwanted flow
 */
static inline int parse_header(unsigned char *buf, dns_record_t *dns, int buflen)
{
	PRINTD("DNS header parsing ...\n");
	unsigned char *position = buf;	/* position in packet data buffer */
	size_t off = DNS_OFFSET;	/* first offset */
	uint16_t tmp;
	
	dns_detail_t dns_detail;
	memset(&dns_detail, 0, sizeof(dns_detail_t));
	
	/* parse DNS header ID and FLAGS */
	memcpy(&(tmp), position, sizeof(uint16_t));
	dns->id = ntohs(tmp);
	
	memcpy(&(dns_detail.flags), position + sizeof(uint16_t), sizeof(uint16_t));
;
	/* parse DNS header number of records of each type */
	memcpy(&(tmp), position + sizeof(uint16_t)*2, sizeof(uint16_t));
	dns_detail.questions = ntohs(tmp);
	memcpy(&(tmp), position + sizeof(uint16_t)*4, sizeof(uint16_t));
	dns_detail.authority = ntohs(tmp);
	memcpy(&(tmp), position + sizeof(uint16_t)*5, sizeof(uint16_t));
	dns_detail.additional = ntohs(tmp);
	memcpy(&(tmp), position + sizeof(uint16_t)*3, sizeof(uint16_t));
	dns->answers = ntohs(tmp);

	/* parse every flag from FLAGS sequence */
	uint8_t hi = dns_detail.flags ;
	uint8_t lo = dns_detail.flags >> 8;
	  
	dns_detail.qr = ((hi & 0x80) == 0x80) ? 0x1 : 0x0;
	dns->rcode = (lo & 0xf);
	if (ALL_DNS) {
		dns_detail.opcode = (hi & 0x78);
		dns_detail.aa = ((hi & 0x4) == 0x4) ? '1' : '0';
		dns_detail.tc = ((hi & 0x2) == 0x2) ? '1' : '0';
		dns_detail.rd = ((hi & 0x1) == 0x1) ? '1' : '0';
		dns_detail.ra = ((lo & 0x80) == 0x80) ? '1' : '0';
		dns_detail.z = ((lo & 0x40) == 0x40)  ? '1' : '0';
		dns_detail.ad = ((lo & 0x0020) == 0x0020) ? '1' : '0';
		dns_detail.cd = ((lo & 0x0010) == 0x0010) ? '1' : '0';
	}	
		
	/* packet validity */
	dns_detail.valid = 1;
	
#ifdef PARSE_ALL
	dns->d = dns_detail;
#endif
	
	/* query parsing */
	if ( ERR == parse_query(position, buflen, &off, dns))
		return ERR;
	
	/* response parsing */
	if (dns->rcode == 0) {
		if (dns_detail.qr == 1) {
			if (dns->answers == 0) {
				/* server is not authoritative for query, authority/additional section is parsed */
			} else {
				/* server is authoritative for query, answers section is parsed */
				if ( ERR == parse_record(position, buflen, &off, dns)) {
					/* response is empty or invalid*/
					strcpy(dns->rdata, "(empty)");
					return ERR;
				}
				
				if (dns->answers > 1) {
					int i;
					for (i = 2; i <= dns->answers; i++)
					      parse_skip(position, buflen, &off);
				}
			}
		}
		else {
			/* its query, not response */
			strcpy(dns->rdata, "(query)");
		}
		
	} else {
		/* RCODE is not 0, answer is not valid, see list of RCODEs */
		strcpy(dns->rdata, "(error)");
	}

	if (dns_detail.authority > 0) {
		int i;
		for (i = 1; i <= dns_detail.authority; i++)
			parse_skip(position, buflen, &off);
	}
	
	if (dns_detail.additional > 0) {
		int i;
		for (i = 1; i <= dns_detail.additional; i++)
			parse_additional(position, buflen, &off, dns);
	}
	PRINTD("S: %d ", (int)sizeof(dns_record_t));

	return OK;
}

/* USED LENGTHS OF VALUES */
/**
 * \brief function for specifying length of getter
 * @return sizeof(uint8_t)
 */
int value_length_uint8(void *self, flow_record_t *r)
{
	return sizeof(uint8_t);
}

/**
 * \brief function for specifying length of getter
 * @return sizeof(uint16_t)
 */
int value_length_uint16(void *self, flow_record_t *r)
{
	return sizeof(uint16_t);
}

/**
 * \brief function for specifying length of getter
 * @return sizeof(uint32_t)
 */
int value_length_uint32(void *self, flow_record_t *r)
{
	return sizeof(uint32_t);
}

/**
 * \brief function for specifying length of getter
 * @return sizeof(char)
 */
int value_length_char(void *self, flow_record_t *r)
{
	return sizeof(char);
}

/**
 * \brief function for specifying length of getter
 * @return r_data length
 */
int value_length_qlength(void *self, flow_record_t *r)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(r, p->data);

	return strlen(pr->name);
}

/**
 * \brief function for specifying length of getter
 * @return r_data length
 */
int value_length_rlength(void *self, flow_record_t *r)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(r, p->data);

	return pr->rlength;
}

/**
 * \brief function for specifying length of getter
 * @return r_data length
 */
int value_length_slength(void *self, flow_record_t *r)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(r, p->data);

	return strlen(pr->rdata);
}

/* VALIDERS */
/**
 * \brief validity function used to validate getters, always return valid;
 *
 * @return 1 - always valid
 */
int validity_OK(void *self, flow_record_t *r)
{
	return 1;
}

/**
 * \brief check if data in flow record are valid, used to decide, if we want to overwrite flow data
 * @return value of valid field in flow record (0 or 1)
 */
int validity_by_valid(void *self, flow_record_t *r)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(r, p->data);
	
	// dns type based check
	if (pr->type > 0)
	    return 1;
	else
	    return 0;
}

/**
 * \brief check if data in flow record, additional part are valid, used to decide, if we want to overwrite flow data
 * @return value of valid field in flow additional record (0 or 1)
 */
int validity_by_valid_additional(void *self, flow_record_t *r)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(r, p->data);
	
	if (pr->payload_size > 0)
	    return 1;
	else
	    return 0;
}

/* GETTER FILLERS */

/**
 * \brief Function fills dst pointer with DNS id data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_dnsid(void *self, flow_record_t *record, void *dst,
			  int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->id), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS id data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_dnsid_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->id), 0, sizeof(uint16_t), len);
}

#ifdef PARSE_ALL
 /**
 * \brief Function fills dst pointer with DNS flags data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_flags(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.flags), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS flags data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/ 
void value_fill_flags_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.flags), 0, sizeof(uint16_t), len);
}

 /**
 * \brief Function fills dst pointer with DNS number of questions data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/ 
void value_fill_questions(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.questions), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS number of questions data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_questions_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.questions), 0, sizeof(uint16_t), len);
}
#endif

 /**
 * \brief Function fills dst pointer with DNS number of answers data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_answers(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->answers), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS number of answers data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_answers_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->answers), 0, sizeof(uint16_t), len);
}

#ifdef PARSE_ALL
 /**
 * \brief Function fills dst pointer with DNS number of authority records data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_authority(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.authority), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS number of authority records data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_authority_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.authority), 0, sizeof(uint16_t), len);
}


 /**
 * \brief Function fills dst pointer with DNS number of additional records data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_additional(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.additional), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS number of additional records data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_additional_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.additional), 0, sizeof(uint16_t), len);
}


 /**
 * \brief Function fills dst pointer with DNS query/response flag data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_qr(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.qr), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS query/response flag data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_qr_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.qr), 0, sizeof(uint8_t), len);
}


/**
 * \brief Function fills dst pointer with DNS opcode data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_opcode(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.opcode), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS opcode data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_opcode_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.opcode), 0, sizeof(uint8_t), len);
}


 /**
 * \brief Function fill dst pointer with DNS authoritative answer flag data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_aa(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.aa), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS authoritative answer flag data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_aa_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.aa), 0, sizeof(char), len);
}


 /**
 * \brief Function fills dst pointer with DNS truncation flag data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_tc(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.tc), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS truncation flag data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_tc_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.tc), 0, sizeof(char), len);
}


 /** 
 * \brief Function fills dst pointer with DNS recursion desired flag data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_rd(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.rd), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS recursion desired flag data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_rd_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.rd), 0, sizeof(char), len);
}


 /**
 * \brief Function fills dst pointer with DNS recursion available data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_ra(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.ra), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS recursion available data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_ra_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.ra), 0, sizeof(char), len);
}


 /**
 * \brief Function fills dst pointer with DNS Z bit data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_z(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.z), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS Z bit data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_z_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.z), 0, sizeof(char), len);
}


 /**
 * \brief Function fills dst pointer with DNS authenticated data flag data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_ad(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.ad), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS authenticated data flag data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_ad_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.ad), 0, sizeof(char), len);
}


 /**
 * \brief Function fills dst pointer with DNS checking disabled flag data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
*/
void value_fill_cd(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->d.cd), 0, sizeof(char), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS checking disabled flag data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
*/
void value_fill_cd_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->d.cd), 0, sizeof(char), len);
}
#endif

/**
 * \brief Function fills dst pointer with DNS return code data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_rcode(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->rcode), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS recturn code data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_rcode_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->rcode), 0, sizeof(uint8_t), len);
}


 /**
 * \brief Function fills dst pointer with DNS hostname data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_name(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);


	record_universal_get(dst, &(pr->name), 0, strlen(pr->name), len, 0);
}

/**
 * \brief Function fills dst pointer with txt DNS hostname data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_name_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->name), 0, strlen(pr->name), len, 0);
}


/**
 * \brief Function fills dst pointer with DNS query type data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_querytype(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);
 
	record_universal_get(dst, &(pr->type), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS query type data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_querytype_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->type), 0, sizeof(uint16_t), len);
}


/**
 * \brief Function fills dst pointer with DNS class data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_class(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);
 
	record_universal_get(dst, &(pr->class), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS class data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_class_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->class), 0, sizeof(uint16_t), len);
}


/**
 * \brief Function fills dst pointer with DNS response TTL data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_rrttl(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);
 
	record_universal_get(dst, &(pr->ttl), 0, sizeof(uint32_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS response TTL data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_rrttl_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->ttl), 0, sizeof(uint32_t), len);
}


/**
 * \brief Function fills dst pointer with DNS resource record length data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_rlength(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);
 
	record_universal_get(dst, &(pr->rlength), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS resource record length data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_rlength_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->rlength), 0, sizeof(uint16_t), len);
}


 /**
 * \brief Function fills dst pointer with DNS resource record data
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_rdata(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->rdata), 0, strlen(pr->rdata), len, 0);
}

/**
 * \brief Function fills dst pointer with txt DNS resource record data
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_rdata_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->rdata), 0, strlen(pr->rdata), len, 0);
}


 /**
 * \brief Function fills dst pointer with DNS additional requested UDP record size
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_psize(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->payload_size), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS additional requested UDP record size
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_psize_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->payload_size), 0, sizeof(uint16_t), len);
}


 /**
 * \brief Function fills dst pointer with DNS additional DNSKEY OK bit
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @param to_network_byte_order specify byte order, set to 0 or 1
 * @return void
 */
void value_fill_do(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get(dst, &(pr->do_bit), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Function fills dst pointer with txt DNS additional DNSKEY OK bit
 * \param *self pointer to plugin private structure
 * \param *record pointer to location to copy data
 * \param *dst Where to copy resulted data
 * \param len length of getter (0 for non-constant length)
 */
void value_fill_do_txt (void *self, flow_record_t *record, char *dst, int len)
{
	plugin_private_dns *p = (plugin_private_dns *) self;
	dns_record_t *pr = PLUGIN_DATA(record, p->data);

	record_universal_get_txt_num (dst, &(pr->do_bit), 0, sizeof(uint8_t), len);
}

 /**
 * \brief Dummy filler, fills the dst with nulls
 * @param *self pointer to plugin private structure
 * @param *record pointer to flow data
 * @param *dst pointer to location to copy data
 * @param len length of getter (0 for non-constant length)
 * @return void
 */
void value_fill_dummy_txt(void *self, flow_record_t *record, char *dst, int len)
{
	memset(dst, '\0', len);
}

/* MAIN PLUGIN FUNCTIONS */
/**
 * \brief initialization of input-dns plugin
 * @param *params character field
 * @param getter_list List of getters, used for getter initialization
 * @param data_offset Offset to acces plugin's private structure
 * @param queue_count Number of used queues
 * @return pointer to string with filter
 */
void *plugin_process_init(char *params, flow_record_getter_t **getter_list, int data_offset, int queue_count)
{
	/* Set verbose, debug mode */
	VERBOSE = (verbose_level >= 0) ? true : false;
	DBG = (debug_level >= 0) ? true : false;

	/* Allocate private structure */
	plugin_private_dns *retval = malloc(sizeof(plugin_private_dns));
	if (retval == NULL) {
		return NULL;
	}

	/* Init input plugin private config structure */
	PRINTD("DNS input-filter plugin init arguments: %s\n", params);

	retval->data = data_offset;

	/* Parse command line parameters */
	if (parse_params(params, retval) != 0) {
		return NULL;
	}
	
	/* OUTPUT FIELDS SPECIFICATION */
	
	getter_add(getter_list, "DNS_ID", sizeof(uint16_t), retval, 
		    &validity_by_valid, &value_length_uint16, &value_fill_dnsid, &value_fill_dnsid_txt);
#ifdef PARSE_ALL
	if (ALL_DNS) {
	  getter_add(getter_list, "DNS_FLAGS", sizeof(uint16_t), retval,
		    &validity_by_valid, &value_length_uint16, &value_fill_flags, &value_fill_flags_txt);
	  getter_add(getter_list, "DNS_QUESTIONS", sizeof(uint16_t), retval, 
		    &validity_by_valid, &value_length_uint16, &value_fill_questions, &value_fill_questions_txt);
	}
#endif
	getter_add(getter_list, "DNS_ANSWERS", sizeof(uint16_t), retval,
		   &validity_by_valid, &value_length_uint16, &value_fill_answers, &value_fill_answers_txt);
#ifdef PARSE_ALL
	if (ALL_DNS) {
	  getter_add(getter_list, "DNS_AUTHORITY", sizeof(uint16_t), retval,
		    &validity_by_valid, &value_length_uint16, &value_fill_authority, &value_fill_authority_txt);
	  getter_add(getter_list, "DNS_ADDITIONAL", sizeof(uint16_t), retval, 
		    &validity_by_valid, &value_length_uint16, &value_fill_additional, &value_fill_additional_txt);
	  getter_add(getter_list, "DNS_QR", sizeof(uint8_t), retval, 
		    &validity_by_valid, &value_length_uint8, &value_fill_qr, &value_fill_qr_txt);
	  getter_add(getter_list, "DNS_OPCODE", sizeof(uint8_t), retval, 
		    &validity_by_valid, &value_length_uint8, &value_fill_opcode, &value_fill_opcode_txt);
	  getter_add(getter_list, "DNS_AA", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_aa, &value_fill_aa_txt);
	  getter_add(getter_list, "DNS_TC", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_tc, &value_fill_tc_txt);
	  getter_add(getter_list, "DNS_RD", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_rd, &value_fill_rd_txt);
	  getter_add(getter_list, "DNS_RA", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_ra, &value_fill_ra_txt);
	  getter_add(getter_list, "DNS_Z", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_z, &value_fill_z_txt);
	  getter_add(getter_list, "DNS_AD", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_ad, &value_fill_ad_txt);
	  getter_add(getter_list, "DNS_CD", sizeof(char), retval, 
		    &validity_by_valid, &value_length_char, &value_fill_cd, &value_fill_cd_txt);
	}
#endif
	getter_add(getter_list, "DNS_RCODE", sizeof(uint8_t), retval, 
		   &validity_by_valid, &value_length_uint8, &value_fill_rcode, &value_fill_rcode_txt);
	getter_add(getter_list, "DNS_NAME", -1, retval, 
		   &validity_by_valid, &value_length_qlength, &value_fill_name, &value_fill_name_txt);
	getter_add(getter_list, "DNS_QTYPE", sizeof(uint16_t), retval, 
		   &validity_by_valid, &value_length_uint16, &value_fill_querytype, &value_fill_querytype_txt);
	getter_add(getter_list, "DNS_CLASS", sizeof(uint16_t), retval, 
		   &validity_by_valid, &value_length_uint16, &value_fill_class, &value_fill_class_txt);
	getter_add(getter_list, "DNS_RR_TTL", sizeof(uint32_t), retval, 
		   &validity_by_valid, &value_length_uint32, &value_fill_rrttl, &value_fill_rrttl_txt);
	getter_add(getter_list, "DNS_RLENGTH", sizeof(uint16_t), retval, 
		   &validity_by_valid, &value_length_uint16, &value_fill_rlength, &value_fill_rlength_txt);
	getter_add(getter_list, "DNS_RDATA", -1, retval, 
		&validity_by_valid, &value_length_slength, &value_fill_rdata, &value_fill_rdata_txt);
	getter_add(getter_list, "DNS_PSIZE", sizeof(uint16_t), retval, 
		&validity_by_valid_additional, &value_length_uint16, &value_fill_psize, &value_fill_psize_txt);
	getter_add(getter_list, "DNS_DO", sizeof(uint8_t), retval, 
		&validity_by_valid_additional, &value_length_uint8, &value_fill_do, &value_fill_do_txt);

	return retval;
}

/**
 * \brief Parse packet and assign it to flow based on the hash
 * @param *plugin_private plugin's private structure
 * @param *packet packet full structure with pointers and length
 * @return FLOW_FILTER_PASS or FLOW_FILTER_DROP
 */
int plugin_process_filter(void *plugin_private, flow_packet_t *packet)
{
	plugin_private_dns *dev = (plugin_private_dns *) plugin_private;
	dns_record_t *precord = PLUGIN_DATA(packet->record, (dev->data));
	
	/* initialize record */
	memset((void *)precord, 0, sizeof(dns_record_t));

	/* process only UDP packets */
	if (!(packet->record->l4.protocol == UDP)) {
	    return FLOW_FILTER_PASS;
	}

	/* process only DNS traffic */
	if (((packet->record->l4.src_port) != DNS) && ((packet->record->l4.dst_port) != DNS)) {
	    return FLOW_FILTER_PASS;
	}
	
	/* packet parameters debug print */
	//PRINTD("P.start:%d,P.current:%d,P.length:%d,L3.off:%d,L4.off:%d\n", packet->packet_start, packet->packet_current, packet->packet_len, packet->l3_offset, packet->l4_offset);
	
	/* parse DNS packet header and payload */
	if (ERR == parse_header(packet->packet_current, precord, (packet->packet_len - (packet->l4_offset + UDP_SIZE)) ) )
	  return FLOW_FILTER_PASS;
	
	return FLOW_FILTER_PASS;
}

/**
 * \brief Function determining behavior on flow update
 *
 * @param *plugin_private plugin's private structure
 * @param *record pointer to flow record generated by past packets
 * @param *update pointer to flow packet data generated by new packet
 * @return void
 */
void plugin_process_update(void *plugin_private, flow_record_t *record, flow_packet_t *update)
{	
	plugin_private_dns *dev = (plugin_private_dns *) plugin_private;
	dns_record_t *original = PLUGIN_DATA(record, dev->data);
	dns_record_t *up = PLUGIN_DATA(update->record, dev->data);

	if (up->type > 0) {
		if (original->type > 0) {
			/* SPLIT FLOW - every packet has its own flow */
			record->flush = FLOW_FLUSH_SPLIT;
		}
	}
}


/**
 * \brief Function determining behavior on the queue initialization 
 * is obligatory
 *
 * @param *base_plugin_private plugin data
 * @param queue_id id of queue
 * @return plugin private structure
 */
void *plugin_process_init_queue(void *base_plugin_private, int queue_id)
{
	/* Nothing to do */
	return base_plugin_private;
}

/**
 * \brief Function determining behavior on flow release is obligatory
 *
 * @param *plugin_private plugin's private structure
 * @param *record pointer to flow record generated by past packets
 * @param reason release reason code
 * @return void
 */
void plugin_process_release(void *plugin_private, flow_record_t *record, int reason)
{
	/* Nothing to do */
}

/**
 * \brief Plugin post-process function is obligatory
 *
 * @param *plugin_private Plugin private data
 * @param *record Flow record
 * @return void
 */
void plugin_process_postprocess(void *plugin_private, flow_record_t *record)
{
	/* Nothing to do */
}

/**
 * \brief Plugin shutdown function is obligatory
 *
 * @param *plugin_private Plugin private data
 * @return zero
 */
int plugin_process_shutdown(void *plugin_private)
{
	/* Nothing to do */
	free(plugin_private);
	return 0;
}
