/**
 * \file dns.h
 * \author Michal Kovacik <ikovacik@fit.vutbr.cz>#
 * \date 02.02.2014
 * \brief Process plugin for parsing DNS headers on input in 
 * flowmonexp 4.0 and later, header file
 *
 * Copyright (C) 2014 CESNET
 *
 * LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL) version 2 or later, in which case the provisions
 * of the GPL apply INSTEAD OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id$
 *
 */

#ifndef PLUGIN_FLOW_DEFINES_H_INCLUDED
#define PLUGIN_FLOW_DEFINES_H_INCLUDED

#define DNS_TYPE_A 1
#define DNS_TYPE_NS 2
#define DNS_TYPE_CNAME 5
#define DNS_TYPE_SOA 6
#define DNS_TYPE_PTR 12
#define DNS_TYPE_HINFO 13
#define DNS_TYPE_MINFO 14
#define DNS_TYPE_MX 15
#define DNS_TYPE_TXT 16
#define DNS_TYPE_ISDN 20
#define DNS_TYPE_AAAA 28
#define DNS_TYPE_SRV 33
#define DNS_TYPE_DNAME 39
#define DNS_TYPE_OPT 41
#define DNS_TYPE_DS 43
#define DNS_TYPE_RRSIG 46
#define DNS_TYPE_DNSKEY 48

#define DNS_CLASS_INTERNET 1
#define DNS_CLASS_CHAOS 3
#define DNS_CLASS_HESIOD 4

#define DNS_OPCODE_QUERY 0
#define DNS_OPCODE_INVERSE_QUERY 1
#define DNS_OPCODE_STATUS 2

#define DNS_RETURN_CODE_NO_ERROR 0
#define DNS_RETURN_CODE_FORMAT_ERROR 1
#define DNS_RETURN_CODE_SERVER_FAILURE 2
#define DNS_RETURN_CODE_NAME_ERROR 3
#define DNS_RETURN_CODE_NOT_IMPLEMENTED 4
#define DNS_RETURN_CODE_REFUSED 5
#define DNS_RETURN_CODE_YXDOMAIN 6
#define DNS_RETURN_CODE_YXRRSET 7
#define DNS_RETURN_CODE_NXRRSET 8
#define DNS_RETURN_CODE_NOT_AUTH 9
#define DNS_RETURN_CODE_NOT_ZONE 10

#define DNS_OFFSET 12			/** DNS header size offset*/
#define DNS_NAME_COMPRESSED 192		/** 0xc0 for hostnames with pointer */
#define DNS_NAME_UNCOMPRESSED 64	/** full hsotname without pointer */
#define DNS_MAXDEPTH 32			/** max depth of parse name recursion */
#define DNS_IP6CHARS 16			/** number of 4bit chars for IPv6 */

/* IMPORTANT CONSTANT SETTINGS */
#define RESPONSE_SIZE 160		/** max rdata size */
#define NAME_SIZE 128			/** max name size */
#define LEVEL_SIZE 128			/** size of each domain level */

#define DNS 53				/** DNS port */
#define TCP 6				/** TCP protocol number */
#define UDP 17				/** UDP protocol number */
#define UDP_SIZE 8			/** L4 UDP size */
#define STRING_CUT_SIZE 41		/** fixed length char* variable size */
#define BASE 10				/** base for number conversion **/

#define ERR -1				/** error return code */
#define OK 0				/** ok return code */

/** structure for query data with fixed size */
typedef struct query_line_st
{
	uint16_t type;
	uint16_t class;
}__attribute__((__packed__))query_line_t;
 
/** structure for response record data with fixed size */
typedef struct record_line_st
{
	uint16_t type;
	uint16_t class;
	uint32_t ttl;
	uint16_t data_len;
}__attribute__((__packed__))rr_line_t;

/** structure for soa record type data with fixed size */
struct soa_data_t
{
	uint32_t serial;
	uint32_t refresh;
	uint32_t retry;
	uint32_t expire;
	uint32_t minimum;
}__attribute__((__packed__));

/** structure for srv record type data with fixed size */
struct srv_data_t
{
	uint16_t prio;
	uint16_t weight;
	uint16_t port;
}__attribute__((__packed__));

/** structure for ds record type data with fixed size */
struct ds_data_t
{
	uint16_t keytag;
	uint8_t alg;
	uint8_t digest;
}__attribute__((__packed__));

/** structure for rrsig record type data with fixed size */
struct rrsig_data_t
{
	uint16_t type;
	uint8_t alg;
	uint8_t labels;
	uint32_t ttl;
	uint32_t sig_exp;
	uint32_t sig_inc;
	uint16_t keytag;
}__attribute__((__packed__));

/** structure for dnskey record type data with fixed size */
struct dnskey_data_t
{
	uint16_t flags;
	uint8_t proto;
	uint8_t alg;
}__attribute__((__packed__));

/** query full structure */
typedef struct query_st
{
	/** full query name */
	char name[NAME_SIZE];
	uint16_t type;
	uint16_t class;
	/** length of qname formatted string */
	uint16_t qlength;
}__attribute__((__packed__))query_t;

/** MX response record type full structure */
struct MxRecord_t
{
	uint16_t priority;
	/** mx hostname */
	char mxhost[NAME_SIZE];
}__attribute__((__packed__));
 
/** SRV response record type structure */
struct SrvRecord_t
{
	char service[NAME_SIZE/8];
	char protocol[NAME_SIZE/8];
	char name[NAME_SIZE/2];
	char target[NAME_SIZE];
	uint16_t priority;
	uint16_t weight;
	uint16_t port;
}__attribute__((__packed__));

/** SOA response record type structure */
struct SoaRecord_t
{
	/** nameserver for the zone which holds the zone files */
	char mname[NAME_SIZE/2];
	/** email on an domain administrator */
	char rname[NAME_SIZE/2];
	uint32_t serial;
	uint32_t refresh;
	uint32_t retry;
	uint32_t expire;
	uint32_t min_ttl;
}__attribute__((__packed__));

/** DS respnse record type structure */
struct DsRecord_t
{
	uint16_t keytag;
	uint8_t algorithm;
	uint8_t digest;
	char publickey[NAME_SIZE];
}__attribute__((__packed__));

/** RRSIG response record type structure */
struct RrsigRecord_t
{
	uint16_t type_covered;
	uint8_t algorithm;
	uint8_t labels;
	uint32_t original_ttl;
	uint32_t signature_expiration;
	uint32_t signature_inception;
	uint16_t keytag;
	char signer_signature[NAME_SIZE];
}__attribute__((__packed__));

/** DNSKEY response record type structure */
struct DnsKeyRecord_t
{
	uint16_t flags;
	uint8_t protocol; /* should be 3 */
	uint8_t algorithm;
	char publickey[NAME_SIZE];
}__attribute__((__packed__));

/** RAW response record type structure */
struct RawRecord_t
{
	size_t data_len;
	/** raw data */
	char data[RESPONSE_SIZE];
}__attribute__((__packed__));

/** response record full structure */
typedef struct record_st
{
	uint32_t ttl;
	uint16_t type;
	uint16_t class;
	uint16_t rlength;
	/** length of rdata formatted string */
	uint16_t slength;
	/** full response data */
	char rdata[RESPONSE_SIZE];
	/** full response name */
	char name[NAME_SIZE];

	/** type of response data */
	union
	{
	  char hostname[NAME_SIZE];
	  struct SoaRecord_t soa;
	  struct MxRecord_t mx;
	  struct SrvRecord_t srv;
	  struct DsRecord_t ds;
	  struct DnsKeyRecord_t dnskey;
	  struct RrsigRecord_t rrsig;
	  struct RawRecord_t raw;
	} data;
		
}__attribute__((__packed__))record_t;

/** additional record full structure */
typedef struct additional_st
{
	/** full response name */
	char name[NAME_SIZE];
	/** type of additional record */
	uint16_t type;
	/** requestors UDP payload size */
	uint16_t payload_size;
	/** extended rcode and flags */
	uint32_t flags;
	/** extended RCODE */
	uint8_t ext_rcode;
	/** indicates implementation level of setter */
	uint8_t version;
	/** DNSKEY OK bit */
	uint8_t do_bit;
	/** for now set 0 by senders and ignored by receivers */
	uint16_t z_bit;
	/** length of all rdata */
	uint16_t rlength;
	/** variable length part NOT PARSED */
	//char *rdata;
	/** validity */
	uint8_t add_valid;
		
}__attribute__((__packed__))additional_t;

/** Structure keeps all parsed DNS data (ONLY NOT EXPORTED) */
typedef struct dns_detail_st {
  
	/** Whole flags structure */
	uint16_t flags;
	/** QR, Query/Response. 1 bit. QR Description: 0=Query,1=Response. */
	uint8_t qr;
	/** Opcode. 4 bits. */
	uint8_t opcode;
	/** AA, Authoritative Answer. 1 bit. 0=Not authoritative,1=Is authoritative. */
	char aa;
	/** TC, Truncated. 1 bit. Indicates that only the first 512 bytes of the reply was returned. TC	Description 0=Not truncated,1=Message truncated. */
	char tc;
	/** RD, Recursion Desired. 1 bit. RD Description: 0=Recursion not desired,1=Recursion desired. */
	char rd;
	/** RA, Recursion Available. 1 bit.0=Recursive query support not available,1=Recursive query support available. */
	char ra;
	/** Z. 1 bit. */
	char z;
	/** AD, Authenticated data. 1 bit. */
	char ad;
	/** CD, Checking Disabled. 1 bit. */
	char cd;
	/** Total Questions. 16 bits, unsigned.Number of entries in the question list that were returned. */
	uint16_t questions;
	/** Total Authority RRs. 16 bits, unsigned.Number of entries in the authority resource record list that were returned. */
	uint16_t authority;
	/** Total Additional RRs. 16 bits, unsigned.Number of entries in the additional resource record list that were returned. */
	uint16_t additional;
	/* validity */
	uint8_t valid;

	/** query data */
	query_t q;
	/** response data */
	record_t r;
	/** additional data */
	additional_t a;
	
}__attribute__((__packed__))dns_detail_t;

/** Structure keeps all DNS data for plugin defined getters (ONLY EXPORTED) */
/* ORDER IS IMPORTANT DUE TO MEMORY ALIGNATION !!! */
typedef struct dns_record_st {
	
	/** response ttl */
	uint32_t ttl;
	/** Identification. 16 bits. Used to match request/reply packets. */
	uint16_t id;
	/** Total Answer RRs. 16 bits, unsigned.Number of entries in the answer resource record list that were returned. */
	uint16_t answers;
	/** query type */
	uint16_t type;
	/** query class */
	uint16_t class;
	/** response length */
	uint16_t rlength;
	/** requestors UDP payload size */
	uint16_t payload_size;
	/** DNSKEY OK bit */
	uint8_t do_bit;
	/** Rcode, Return code. 4 bits. */
	uint8_t rcode;
	/** full response data */
	char rdata[RESPONSE_SIZE];
	/** query name */
	char name[NAME_SIZE];

#ifdef PARSE_ALL
	dns_detail_t d;
#endif
	
}__attribute__((__packed__))dns_record_t;

/** flowmonexp verbose level*/
extern int verbose_level;

/** flowmonexp debug level*/
extern int debug_level;


#endif /* PLUGIN_FLOW_DEFINES_H_INCLUDED */
